@@ -545,7 +545,11 @@ class ModelSet:
         period_name="Q",
         extract_drivers=True,
         historical_only=False,
+        allow_nulls=False,
+        mo_only=False,
     ):
+        self.allow_nulls = allow_nulls
+        self.mo_only = mo_only
         self.config = config or CONFIG
         self.historical_only = historical_only
         self.period_name = period_name
@@ -660,14 +664,18 @@ class ModelSet:
                     list_all.append(list_rows)
 
             df = pd.concat(list_df)
-            # df = df.sort_values(["period_name", "name_index"])
 
-            res = list(reduce(lambda i, j: i & j, (set(x) for x in list_all)))
-            self.common_time_series = res
-
-            df = df_filter(df, "time_series_name", res)
-
-            self._features = df
+            if self.allow_nulls == False:
+                res = list(reduce(lambda i, j: i & j, (set(x) for x in list_all)))
+                self._common_time_series_names = res
+                df = df_filter(df, "time_series_name", res)
+                self._features = df
+            else:
+                if self.mo_only == True:
+                    df = df.loc[df["time_series_name"].str.startswith("MO_")]
+                    self._features = df
+                else:
+                    self._features = df
 
             if self.extract_drivers == True:
                 self.drivers = df[df["is_driver"] == 1]
@@ -1263,6 +1271,71 @@ class Model:
             print("guidance")
             self.create_guidance_csv()
 
+    def key_driver_map(self, time_series_name):
+        # defaulting to MO_RIS_REV, but you could choose another time_series_name as a starting point like MO_RIS_EBIT
+        model_map = self.create_model_map(
+            time_series_name=time_series_name, col_for_labels="time_series_name"
+        )
+
+        # list all the nodes in the model map
+        model_map_time_series_list = model_map.list_time_series()
+
+        # make it a dataframe because I'm a reformed R user
+        df = pd.DataFrame(model_map_time_series_list)
+
+        # columns for joining
+        df.columns = ["time_series_name"]
+        df["ticker"] = self.ticker
+
+        # create a model frame with only key drivers and just the most recent quarter
+        df_drivers = self.model_frame(
+            is_driver=True, period_duration_type="fiscal_quarter", mrq=True
+        )
+        # merge (inner join)
+        df = pd.merge(
+            df,
+            df_drivers,
+            how="inner",
+            left_on=["ticker", "time_series_name"],
+            right_on=["ticker", "time_series_name"],
+        )
+
+        return df
+
+    def forecast_frame(
+        self, time_series_name, n_periods, function_name="value", function_value=""
+    ):
+        df = self.model_frame(
+            [time_series_name],
+            period_duration_type="fiscal_quarter",
+            is_historical=False,
+            n_periods=n_periods,
+        )
+
+        if function_name != "value":
+            d = FuncMusic()
+            df["new_value"] = df.apply(
+                lambda row: d.apply_function(
+                    row["value"], modifier=function_name, argument=function_value
+                ),
+                axis=1,
+            )
+        else:
+            df["new_value"] = function_value
+
+        df = df[
+            [
+                "ticker",
+                "period_name",
+                "time_series_name",
+                "value",
+                "new_value",
+                "period_end_date",
+            ]
+        ]
+
+        return df
+
     def create_time_series_chart(self, time_series_name):
         # get needed data from dataframe
         df = self._model_frame[
@@ -1837,31 +1910,23 @@ class Model:
 
         url = get_forecast_url(self.csin, self.latest_version, self.config.mds_host)
 
-        # print(url)
-
         scenario_url = (
             f"{self.config.mds_host}/"
             f"{SCENARIO_URL.format(csin=self.csin, version=self.latest_version)}"
-        )
+        ) + "?page_size=200"
 
         scenario_response = get_request(scenario_url, headers=self.api_headers)
 
-        # print("scenario response"+scenario_response)
-
         scenario_json = scenario_response.json()
-        # print("scenario json"+scenario_json)
 
         scenario_id_url = map_scenario_urls(scenario_json).get(self.uuid)
 
-        # print("uuid"+str(self.uuid))
-        # print("scenario id url: "+str(scenario_id_url))
-
         if scenario_id_url is None:
-            # self.log.write(
+
             print(
                 "Scenario ID for " + scenario_url + " is None.  Perhaps max scenarios."
             )
-            # )
+
             return
 
         print(self.ticker + " scenario_id_url: " + str(scenario_id_url))
